"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const chalk_1 = require("chalk");
const fs = require("fs");
const path = require("path");
const polymer_workspaces_1 = require("polymer-workspaces");
const convert_workspace_1 = require("../convert-workspace");
const githubTokenMessage = `
You need to create a github token and place it in a file named 'github-token'.
The token does not need any permissions.

Generate a token here:   https://github.com/settings/tokens

Then:

echo 'PASTE TOKEN HEX HERE' > ./github-token
`;
/**
 * Checks for github-token in the RunnerOptions and if not specified, will look
 * in the github-token file in working folder.  If that doesn't exist either,
 * we message to the user that we need a token and exit the process.
 */
function loadGitHubToken() {
    // TODO(usergenic): Maybe support GITHUB_TOKEN as an environment variable,
    // since this would be a better solution for Travis deployments etc.
    const githubFilename = 'github-token';
    if (!fs.existsSync(githubFilename)) {
        console.error(`Missing file "${githubFilename}"`);
        return null;
    }
    try {
        return fs.readFileSync(githubFilename, 'utf8').trim();
    }
    catch (e) {
        console.error(`Unable to load file ${githubFilename}: ${e.message}`);
    }
    return null;
}
async function run(options) {
    const workspaceDir = path.resolve(options['workspace-dir']);
    console.log(chalk_1.default.dim('[1/3]') + ' 🚧  ' +
        chalk_1.default.magenta(`Setting Up Workspace "${workspaceDir}"...`));
    if (!options['npm-version']) {
        throw new Error('--npm-version required');
    }
    const npmPackageVersion = options['npm-version'];
    const githubToken = options['github-token'] || loadGitHubToken();
    if (!githubToken) {
        console.log(githubTokenMessage);
        return;
    }
    const workspace = new polymer_workspaces_1.Workspace({
        token: githubToken,
        dir: workspaceDir,
    });
    const reposToConvert = await workspace.init({
        include: options['repo'],
        exclude: options['exclude'],
    }, {
        fresh: options['clean'],
        verbose: true,
    });
    console.log(chalk_1.default.dim('[2/3]') + ' 🌀  ' +
        chalk_1.default.magenta(`Converting ${reposToConvert.length} Package(s)...`));
    await convert_workspace_1.default({
        workspaceDir,
        npmImportStyle: options['import-style'],
        packageVersion: npmPackageVersion,
        reposToConvert,
    });
    console.log(chalk_1.default.dim('[3/3]') + ' 🎉  ' + chalk_1.default.magenta(`Conversion Complete!`));
}
exports.default = run;
//# sourceMappingURL=command-workspace.js.map