"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const polymer_analyzer_1 = require("polymer-analyzer");
const conversion_settings_1 = require("./conversion-settings");
const manifest_converter_1 = require("./manifest-converter");
const project_converter_1 = require("./project-converter");
const special_casing_1 = require("./special-casing");
const package_url_handler_1 = require("./urls/package-url-handler");
const util_1 = require("./urls/util");
const util_2 = require("./util");
/**
 * Create and/or clean the "out" directory, setting it up for conversion.
 */
async function setupOutDir(outDir, clean = false) {
    if (clean) {
        await util_2.rimraf(outDir);
    }
    try {
        await util_2.mkdirp(outDir);
    }
    catch (e) {
        if (e.errno === -17) {
            // directory exists, do nothing
        }
        else {
            throw e;
        }
    }
}
/**
 * Create the default conversion settings, adding any "main" files from the
 * current package's bower.json maifest to the "includes" set.
 */
function getConversionSettings(analysis, options, bowerJson) {
    const conversionSettings = conversion_settings_1.createDefaultConversionSettings(analysis, options);
    let bowerMainFiles = (bowerJson.main) || [];
    if (!Array.isArray(bowerMainFiles)) {
        bowerMainFiles = [bowerMainFiles];
    }
    for (const filename of bowerMainFiles) {
        conversionSettings.includes.add(filename);
    }
    return conversionSettings;
}
/**
 * Get the relevant documents from a package, to be converted.
 */
function getPackageDocuments(analysis, conversionSettings) {
    const htmlDocuments = [...analysis.getFeatures({ kind: 'html-document' })];
    return htmlDocuments.filter((d) => package_url_handler_1.PackageUrlHandler.isUrlInternalToPackage(util_1.getDocumentUrl(d)) &&
        !conversionSettings.excludes.has(d.url));
}
exports.getPackageDocuments = getPackageDocuments;
/**
 * Configure a basic analyzer instance for the package under conversion.
 */
function configureAnalyzer(options) {
    const urlLoader = new polymer_analyzer_1.InMemoryOverlayUrlLoader(new polymer_analyzer_1.FSUrlLoader(options.inDir));
    for (const [url, contents] of special_casing_1.polymerFileOverrides) {
        urlLoader.urlContentsMap.set(url, contents);
        urlLoader.urlContentsMap.set(`bower_components/polymer/${url}`, contents);
    }
    return new polymer_analyzer_1.Analyzer({
        urlLoader,
        urlResolver: new polymer_analyzer_1.PackageUrlResolver(),
    });
}
/**
 * Convert a package-layout project to JavaScript modules & npm.
 */
async function convert(options) {
    const outDir = options.outDir;
    const npmPackageName = options.packageName;
    const npmPackageVersion = options.packageVersion;
    await setupOutDir(outDir, options.cleanOutDir);
    // Configure the analyzer and run an analysis of the package.
    const bowerJson = manifest_converter_1.readJson(options.inDir, 'bower.json');
    const analyzer = configureAnalyzer(options);
    const analysis = await analyzer.analyzePackage();
    await setupOutDir(options.outDir, !!options.cleanOutDir);
    // Create the url handler & converter.
    const urlHandler = new package_url_handler_1.PackageUrlHandler(options.packageName, options.packageType);
    const conversionSettings = getConversionSettings(analysis, options, bowerJson);
    const converter = new project_converter_1.ProjectConverter(urlHandler, conversionSettings);
    // Gather all relevent package documents, and run the converter on them!
    for (const document of getPackageDocuments(analysis, conversionSettings)) {
        converter.convertDocument(document);
    }
    // Filter out external results before writing them to disk.
    const results = converter.getResults();
    for (const [newPath] of results) {
        if (!package_url_handler_1.PackageUrlHandler.isUrlInternalToPackage(newPath)) {
            results.delete(newPath);
        }
    }
    await util_2.writeFileResults(outDir, results);
    // Generate a new package.json, and write it to disk.
    try {
        const packageJson = manifest_converter_1.generatePackageJson(bowerJson, npmPackageName, npmPackageVersion);
        manifest_converter_1.writeJson(packageJson, outDir, 'package.json');
    }
    catch (err) {
        console.log(`error in bower.json -> package.json conversion (${err.message})`);
    }
}
exports.default = convert;
//# sourceMappingURL=convert-package.js.map