"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const mz_1 = require("mz");
const path = require("path");
const polymer_analyzer_1 = require("polymer-analyzer");
const conversion_settings_1 = require("./conversion-settings");
const manifest_converter_1 = require("./manifest-converter");
const project_converter_1 = require("./project-converter");
const special_casing_1 = require("./special-casing");
const workspace_url_handler_1 = require("./urls/workspace-url-handler");
const util_1 = require("./util");
/**
 * Create a symlink from the repo into the workspace's node_modules directory.
 */
async function writeNpmSymlink(options, repo) {
    const packageJsonPath = path.join(repo.dir, 'package.json');
    if (!await mz_1.fs.exists(packageJsonPath)) {
        return;
    }
    const packageJson = manifest_converter_1.readJson(packageJsonPath);
    let packageName = packageJson['name'];
    let parentName = path.join(options.workspaceDir, 'node_modules');
    if (packageName.startsWith('@')) {
        const slashIndex = packageName.indexOf('/');
        const scopeName = packageName.substring(0, slashIndex);
        parentName = path.join(parentName, scopeName);
        packageName = packageName.substring(slashIndex + 1);
    }
    await util_1.mkdirp(parentName);
    const linkName = path.join(parentName, packageName);
    await mz_1.fs.symlink(repo.dir, path.resolve(linkName));
}
/**
 * For a given repo, generate a new package.json and write it to disk.
 */
function writePackageJson(repo, packageVersion) {
    const bowerPackageName = path.basename(repo.dir);
    const bowerJsonPath = path.join(repo.dir, 'bower.json');
    const bowerJson = manifest_converter_1.readJson(bowerJsonPath);
    const npmPackageName = workspace_url_handler_1.lookupNpmPackageName(bowerJsonPath) || bowerPackageName;
    const packageJson = manifest_converter_1.generatePackageJson(bowerJson, npmPackageName, packageVersion);
    manifest_converter_1.writeJson(packageJson, repo.dir, 'package.json');
}
/**
 * Configure a basic analyzer instance for the workspace.
 */
function configureAnalyzer(options) {
    const workspaceDir = options.workspaceDir;
    const urlLoader = new polymer_analyzer_1.InMemoryOverlayUrlLoader(new polymer_analyzer_1.FSUrlLoader(workspaceDir));
    for (const [url, contents] of special_casing_1.polymerFileOverrides) {
        urlLoader.urlContentsMap.set(`polymer/${url}`, contents);
    }
    return new polymer_analyzer_1.Analyzer({
        urlLoader,
        urlResolver: new polymer_analyzer_1.PackageUrlResolver(),
    });
}
/**
 * Convert a set of workspace repos to JavaScript modules & npm.
 */
async function convert(options) {
    const analyzer = configureAnalyzer(options);
    const analysis = await analyzer.analyzePackage();
    const htmlDocuments = [...analysis.getFeatures({ kind: 'html-document' })];
    const conversionSettings = conversion_settings_1.createDefaultConversionSettings(analysis, options);
    const urlHandler = new workspace_url_handler_1.WorkspaceUrlHandler(options.workspaceDir);
    const converter = new project_converter_1.ProjectConverter(urlHandler, conversionSettings);
    // For each repo, convert the relevant HTML documents:
    for (const repo of options.reposToConvert) {
        const repoDirName = repo.dir.split('/').pop();
        const repoDocuments = htmlDocuments.filter((d) => {
            return d.url.startsWith(repoDirName) &&
                !conversionSettings.excludes.has(d.url);
        });
        for (const document of repoDocuments) {
            converter.convertDocument(document);
        }
    }
    // Process & write each conversion result:
    await util_1.writeFileResults(options.workspaceDir, converter.getResults());
    // For each repo, generate a new package.json:
    for (const repo of options.reposToConvert) {
        try {
            writePackageJson(repo, options.packageVersion);
        }
        catch (err) {
            console.log('Error in bower.json -> package.json conversion:');
            console.error(err);
        }
    }
    // For each repo, generate a node_modules/ symlink in the workspace directory:
    for (const repo of options.reposToConvert) {
        try {
            await writeNpmSymlink(options, repo);
        }
        catch (err) {
            console.log(`Error in npm symlink creation:`);
            console.error(err);
        }
    }
}
exports.default = convert;
//# sourceMappingURL=convert-workspace.js.map