"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const manifest_converter_1 = require("../manifest-converter");
const util_1 = require("./util");
/**
 * Handle URLs in a single "package-based" layout. This converter should be used
 * to convert a single package, where all existing Bower dependencies are
 * installed in a "bower_components/" sub-directory inside the main package
 * directory.
 */
class PackageUrlHandler {
    /**
     * Helper function to check if a file URL is internal to the main package
     * being converted (vs. a dependency).
     */
    static isUrlInternalToPackage(url) {
        // OriginalDocumentUrl will always be format `bower_components/*`
        // ConvertedDocument[Url|FilePath] will always be format `./node_modules/*`
        return !url.startsWith('bower_components/') &&
            !url.startsWith('./node_modules/');
    }
    constructor(packageName, packageType = 'element') {
        this.packageName = packageName;
        this.packageType = packageType;
    }
    /**
     * Get the bower package name from an OriginalDocumentUrl and check our
     * dependency map for the matching npm package name match. If URL is internal
     * (no bower package name in URL) return the current package name.
     */
    getPackageNameForUrl(url) {
        // If url contains no directories it must be internal. Explicitly check
        // `basePackageName` here since it is needed below.
        const urlParts = url.split('/');
        const basePackageName = urlParts[1];
        if (!basePackageName || PackageUrlHandler.isUrlInternalToPackage(url)) {
            return this.packageName;
        }
        // For a a single package layout, the Bower package name is included in
        // the URL. ie: bower_components/PACKAGE_NAME/...
        // Check the dependency map to get the new NPM name for the package.
        const depInfo = manifest_converter_1.lookupDependencyMapping(basePackageName);
        if (!depInfo) {
            return basePackageName;
        }
        return depInfo.npm;
    }
    /**
     * Get the "type" for the package where a file lives, based on it's URL.
     */
    getPackageTypeForUrl(url) {
        if (PackageUrlHandler.isUrlInternalToPackage(url)) {
            return this.packageType;
        }
        else {
            return 'element';
        }
    }
    /**
     * Check if two URLs are internal within the same package.
     */
    isImportInternal(fromUrl, toUrl) {
        if (fromUrl.startsWith('./node_modules') &&
            toUrl.startsWith('./node_modules')) {
            return true;
        }
        if (!fromUrl.startsWith('./node_modules') &&
            !toUrl.startsWith('./node_modules')) {
            return true;
        }
        return false;
    }
    /**
     * Rewrite a Bower package name in a URL to its matching npm package name.
     */
    convertUrl(url) {
        if (PackageUrlHandler.isUrlInternalToPackage(url)) {
            // TODO(fks): Revisit this format? The analyzer returns URLs without this
            return ('./' + url);
        }
        const newUrl = url.replace('bower_components/', 'node_modules/');
        const newUrlPieces = newUrl.split('/');
        const bowerPackageName = newUrlPieces[1];
        const depInfo = manifest_converter_1.lookupDependencyMapping(bowerPackageName);
        if (depInfo) {
            newUrlPieces[1] = depInfo.npm;
        }
        return ('./' + newUrlPieces.join('/'));
    }
    /**
     * Get the formatted relative import URL between two ConvertedDocumentUrls.
     */
    getPathImportUrl(fromUrl, toUrl) {
        const isPackageNameScoped = this.packageName.startsWith('@');
        const isPackageTypeElement = this.packageType === 'element';
        const isImportFromLocalFile = PackageUrlHandler.isUrlInternalToPackage(fromUrl);
        const isImportToExternalFile = !PackageUrlHandler.isUrlInternalToPackage(toUrl);
        let importUrl = util_1.getRelativeUrl(fromUrl, toUrl);
        // If the import is from the current project:
        if (isImportFromLocalFile && isPackageTypeElement) {
            // Rewrite imports to point to dependencies as if they were siblings.
            if (importUrl.startsWith('./node_modules/')) {
                importUrl = '../' + importUrl.slice('./node_modules/'.length);
            }
            else {
                importUrl = importUrl.replace('node_modules', '..');
            }
            // Account for a npm package name scoping.
            if (isPackageNameScoped && isImportToExternalFile) {
                if (importUrl.startsWith('./')) {
                    importUrl = '../' + importUrl.slice('./'.length);
                }
                else {
                    importUrl = '../' + importUrl;
                }
            }
        }
        return importUrl;
    }
    /**
     * Get the formatted import URL for a name-based conversion.
     *
     * Ex: `./node_modules/@polymer/polymer/foo.js` -> `@polymer/polymer/foo.js`
     */
    getNameImportUrl(url) {
        return url.slice('./node_modules/'.length);
    }
}
exports.PackageUrlHandler = PackageUrlHandler;
//# sourceMappingURL=package-url-handler.js.map